@@ -27,6 +27,7 @@ from tabulate import tabulate
 from termcolor import colored
 
 from api.main import run_api
+from typhoon.deployment.packaging import typhoon_version_is_local
 from typhoon import connections
 from typhoon.cli_helpers.cli_completion import get_remote_names, get_dag_names, get_conn_envs, get_conn_ids, \
     get_var_types, get_deploy_targets, PROJECT_TEMPLATES, get_task_names
@@ -37,9 +38,10 @@ from typhoon.connections import Connection
 from typhoon.contrib.hooks.hook_factory import get_hook
 from typhoon.core import DagContext
 from typhoon.core.dags import ArgEvaluationError, load_module_from_path
-from typhoon.core.glue import get_dag_errors, load_dag_definition
+from typhoon.core.glue import get_dag_errors, load_dag_definition, load_dag_definitions
 from typhoon.core.settings import Settings
 from typhoon.deployment.packaging import build_all_dags, local_typhoon_path
+from typhoon.deployment.targets.aws import TERRAFORM_FOLDER_PATH
 from typhoon.handler import run_dag
 from typhoon.introspection.introspect_extensions import get_typhoon_extensions, get_typhoon_extensions_info, \
     get_hooks_info
@@ -68,7 +70,8 @@ def set_settings_from_remote(remote: str):
                   file=sys.stderr)
             sys.exit(-1)
         Settings.metadata_db_url = Remotes.metadata_db_url(remote)
-        Settings.fernet_key = Remotes.fernet_key(remote)
+        if Settings.metadata_store(Remotes.aws_profile(remote)).name == 'airflow':
+            Settings.fernet_key = Remotes.fernet_key(remote)
         if Remotes.use_name_as_suffix(remote):
             Settings.metadata_suffix = remote
 
@@ -107,9 +110,14 @@ def init(project_name: str, deploy_target: str, template: str):
             cfg_path = (dest / 'typhoon.cfg')
             dag_schema_path = (dest / 'dag_schema.json')
             component_schema_path = (dest / 'component_schema.json')
-
+        
         cfg_path.write_text(EXAMPLE_CONFIG.format(project_name=project_name, deploy_target=deploy_target))
         Settings.typhoon_home = dest
+        Settings.project_name = project_name
+
+        if template in ['hello_world', 'minimal']:
+            generate_terraform_files(minimal=(template == 'minimal'))
+
         dag_schema, component_schema = generate_json_schemas()
         dag_json_schema = json.dumps(dag_schema, indent=2)
         dag_schema_path.write_text(dag_json_schema)
@@ -251,7 +259,8 @@ def cli_dags():
 @cli_dags.command(name='ls')
 @click.argument('remote', autocompletion=get_remote_names, required=False, default=None)
 @click.option('-l', '--long', is_flag=True, default=False)
-def list_dags(remote: Optional[str], long: bool):
+@click.option('--json-output', is_flag=True, default=False)
+def list_dags(remote: Optional[str], long: bool, json_output):
     set_settings_from_remote(remote)
     metadata_store = Settings.metadata_store(Remotes.aws_profile(remote))
     if long:
@@ -271,11 +280,28 @@ def list_dags(remote: Optional[str], long: bool):
                 ]
                 print(colored(tabulate(table_body, header, 'plain'), 'red'), file=sys.stderr)
     else:
-        for dag_name in sorted(set(x.dag_name for x in metadata_store.get_dag_deployments())):
-            print(dag_name)
-        if not remote:
-            for dag_name, _ in get_dag_errors().items():
-                print(colored(dag_name, 'red'), file=sys.stderr)
+        dag_names = sorted(set(x.dag_name for x in metadata_store.get_dag_deployments()))
+        if json_output:
+            print(json.dumps(dag_names))
+        else:
+            for dag_name in dag_names:
+                print(dag_name)
+            if not remote:
+                for dag_name, _ in get_dag_errors().items():
+                    print(colored(dag_name, 'red'), file=sys.stderr)
+
+
+@cli_dags.command(name='info')
+@click.argument('remote', autocompletion=get_remote_names, required=False, default=None)
+@click.option('--json-output', is_flag=True, default=False)
+@click.option('--indent', type=click.INT, required=False, default=None)
+def dags_info(remote: Optional[str], json_output, indent):
+    set_settings_from_remote(remote)
+    if not json_output:
+        raise NotImplementedError()
+    dags = load_dag_definitions(ignore_errors=True)
+    info = {x.name: {'schedule_interval': x.schedule_interval} for x, _ in dags}
+    print(json.dumps(info, indent=indent))
 
 
 @cli_dags.command(name='build')
@@ -323,10 +349,112 @@ def build_dags(remote: Optional[str], dag_name: Optional[str], all_: bool):
             print(tabulate(table_body, header, 'plain'), file=sys.stderr)
             sys.exit(-1)
         if Settings.deploy_target == 'typhoon':
-            build_all_dags(remote=None, matching=dag_name)
+            build_all_dags(remote=remote, matching=dag_name)
         else:
             from typhoon.deployment.targets.airflow.airflow_build import build_all_dags_airflow
-            build_all_dags_airflow(remote=None, matching=dag_name)
+            build_all_dags_airflow(remote=remote, matching=dag_name)
+
+
+@cli_dags.command(name='push')
+@click.argument('remote', autocompletion=get_remote_names, required=False, default=None)
+@click.option('--dag-name', autocompletion=get_dag_names, required=False, default=None)
+@click.option('--all', 'all_', is_flag=True, default=False, help='Build all DAGs (mutually exclusive with DAG_NAME)')
+@click.option('--code', is_flag=True, default=False, help='Update lambda code but not dependency layer')
+@click.option('--build-deps-locally', is_flag=True, default=False, help='Build dependencies without docker. Not recommended.')
+def push_dags(remote: Optional[str], dag_name: Optional[str], all_: bool, code: bool, build_deps_locally: bool):
+    """Build code for dags in $TYPHOON_HOME/out/"""
+    set_settings_from_remote(remote)
+    if dag_name and all_:
+        raise click.UsageError(f'Illegal usage: DAG_NAME is mutually exclusive with --all')
+    elif dag_name is None and not all_:
+        raise click.UsageError(f'Illegal usage: Need either DAG_NAME or --all')
+    if all_:
+        dag_errors = get_dag_errors()
+        if dag_errors:
+            print(f'Found errors in the following DAGs:')
+            for dag_name in dag_errors.keys():
+                print(f'  - {dag_name}\trun typhoon dag build {dag_name}')
+            sys.exit(-1)
+        raise NotImplementedError()
+    else:
+        if Settings.deploy_target == 'typhoon':
+            build_all_dags(remote=remote, matching=dag_name)
+            dag_out_path = Settings.out_directory/dag_name
+            builds_path = Settings.out_directory/'builds'
+            if builds_path.exists():
+                shutil.rmtree(str(builds_path), ignore_errors=True)
+            builds_path.mkdir()
+            (builds_path/dag_name).mkdir()
+            shutil.make_archive(str(builds_path/dag_name/'lambda'), 'zip', root_dir=str(dag_out_path))
+
+            print('Updating lambda code...')
+            import boto3
+            session = boto3.session.Session(profile_name=Remotes.aws_profile(remote))
+            s3r = session.resource('s3')
+            s3_key_lambda_zip = f'typhoon_dag_builds/{dag_name}/lambda.zip'
+            s3r.Object(Remotes.s3_bucket(remote), s3_key_lambda_zip).put(
+                Body=(builds_path/f'{dag_name}/lambda.zip').open('rb')
+            )
+            lambdac = session.client('lambda')
+            try:
+                lambdac.update_function_code(
+                    FunctionName=f'{dag_name}_{remote}',
+                    S3Bucket=Remotes.s3_bucket(remote),
+                    S3Key=s3_key_lambda_zip,
+                )
+            except lambdac.exceptions.ResourceNotFoundException:
+                print(f'Lambda for {dag_name} does not exist yet. Create it with terraform after this command')
+
+            if not code:
+                print('Building lambda dependencies...')
+                if typhoon_version_is_local():
+                    p = local_typhoon_path().rstrip('typhoon')
+                    local_typhoon_volume = ['-v', f'{p}:{p}']
+                else:
+                    local_typhoon_volume = []
+                if build_deps_locally:
+                    pip_command = f'{sys.executable} -m pip install -r requirements.txt --target .layer/python/'
+                    print(pip_command)
+                    subprocess.run(args=pip_command, cwd=str(dag_out_path), shell=True)
+                else:
+                    docker_pip_command = [
+                        'docker', 'run', '--rm', '-v', f'{dag_out_path}:/var/task',
+                        *local_typhoon_volume,
+                        'lambci/lambda:build-python{}.{}'.format(*sys.version_info),
+                        'pip', 'install', '-r', 'requirements.txt', '--target', './layer/python/',
+                    ]
+                    docker_pip_command = ' '.join(docker_pip_command)
+                    print(docker_pip_command)
+                    subprocess.run(args=docker_pip_command, cwd=str(dag_out_path), shell=True)
+                shutil.make_archive(str(builds_path/dag_name/'layer'), 'zip', root_dir=str(dag_out_path/'layer'))
+                s3_key_layer_zip = f'typhoon_dag_builds/{dag_name}/layer.zip'
+                s3r.Object(Remotes.s3_bucket(remote), s3_key_layer_zip).put(
+                    Body=(builds_path/f'{dag_name}/layer.zip').open('rb')
+                )
+
+                print('Publishing dependencies as layer...')
+                layer_name = f'{dag_name}_dependencies'
+                response = lambdac.publish_layer_version(
+                    LayerName=layer_name,
+                    Description='Dependencies needed for DAG',
+                    Content={
+                        'S3Bucket': Remotes.s3_bucket(remote),
+                        'S3Key': str(s3_key_layer_zip),
+                    },
+                    CompatibleRuntimes=[
+                        'python{}.{}'.format(*sys.version_info)
+                    ]
+                )
+                try:
+                    lambdac.update_function_configuration(
+                        FunctionName=f'{dag_name}_{remote}',
+                        Layers=[response['LayerVersionArn']],
+                    )
+                except lambdac.exceptions.ResourceNotFoundException:
+                    print(f'Lambda for {dag_name} does not exist yet. Create it with terraform after this command')
+        else:
+            raise NotImplementedError()
+
 
 
 @cli_dags.command(name='watch')
@@ -375,8 +503,23 @@ def cli_run_dag(remote: Optional[str], dag_name: str, execution_date: Optional[d
         build_all_dags(remote=None, matching=dag_name)
         run_local_dag(dag_name, execution_date)
     else:
-        # TODO: Run lambda function
-        pass
+        import boto3
+        import base64
+
+        payload_dict = {
+            'type': 'manual',
+            'time': datetime.now().isoformat()
+        }
+        payload = json.dumps(payload_dict).encode()
+
+
+        response = boto3.client('lambda').invoke(
+            FunctionName=f'{dag_name}_{remote}',
+            InvocationType='RequestResponse',
+            LogType='Tail',
+            Payload=payload,
+        )
+        print(base64.b64decode(response['LogResult']).decode())
 
 
 @cli_dags.command(name='definition')
@@ -879,5 +1022,45 @@ Example usage:
 #                 dag_deployment = DagDeployment(loaded_dag['name'], deployment_date=datetime.utcnow(), dag_code=dag_code)
 #                 config.metadata_store.set_dag_deployment(dag_deployment)
 
+@cli.command()
+@click.argument('remote', autocompletion=get_remote_names, required=False, default=None)
+@click.option('-f', '--force', is_flag=True, default=False, help="Force overwrite")
+def generate_terraform(remote: str, force: bool):
+    """Generate terraform files"""
+    set_settings_from_remote(remote)
+    generate_terraform_files(remote)
+
+
+def generate_terraform_files(remote: Optional[str] = None, force: bool = False, minimal: bool = False):
+    terraform_dest_folder = (Settings.typhoon_home/'terraform')
+    if not force and terraform_dest_folder.exists():
+        print(f'Cannot generate terraform files because the folder exists already {terraform_dest_folder}')
+        print('Run with -f/--force to force overwrite')
+        exit(1)
+
+    main_tf_file = TERRAFORM_FOLDER_PATH/'main.tf'
+
+    env_name = remote or 'test'
+    tfvars_template = jinja2.Template((TERRAFORM_FOLDER_PATH/'env.tfvars.j2').read_text())
+    rendered_tfvars = tfvars_template.render(dict(
+        env=env_name,
+        runtime='python{}.{}'.format(*sys.version_info),
+        connections_table=None if minimal else Settings.connections_table_name,
+        variables_table=None if minimal else Settings.variables_table_name,
+        dag_deployments_table=None if minimal else Settings.dag_deployments_table_name,
+        metadata_db_url=None if minimal else Remotes.metadata_db_url(remote),
+        metadata_suffix=None if minimal else Settings.metadata_suffix,
+        s3_bucket=Remotes.s3_bucket(remote) if remote else '',
+        project_name=Settings.project_name,
+    ))
+
+    terraform_dest_folder.mkdir(exist_ok=True)
+    shutil.copy(str(main_tf_file), str(terraform_dest_folder))
+    if not minimal:
+        metadata_store = Settings.metadata_store(Remotes.aws_profile(remote))
+        metadata_store_tf_file = TERRAFORM_FOLDER_PATH/f'metadata_stores/{metadata_store.name}.tf'
+        shutil.copy(str(metadata_store_tf_file), str(terraform_dest_folder))
+    (terraform_dest_folder/f'{env_name}.tfvars').write_text(rendered_tfvars)
+
 if __name__ == '__main__':
     cli()
